from enum import auto, Enum, IntEnum
from typing import TypedDict

from pydantic import BaseModel


class ExarotonResponse(TypedDict):
    success: bool
    error: str
    data: dict | None


class ServerStatus(IntEnum):
    offline = 0
    online = 1
    starting = 2
    stopping = 3
    restarting = 4
    saving = 5
    loading = 6
    crashed = 7
    pending = 8
    transferring = 9
    preparing = 10


class AccountData(BaseModel):
    name: str
    email: str
    verified: bool
    credits: float


class ServerPlayerData(BaseModel):
    max: int
    count: int
    list: list[str]


class ServerSoftwareData(BaseModel):
    id: str
    name: str
    version: str


class ServerData(BaseModel):
    id: str
    name: str
    address: str
    motd: str
    status: ServerStatus
    host: str | None
    port: int | None
    players: ServerPlayerData
    software: ServerSoftwareData | None
    shared: bool


class LogsData(BaseModel):
    content: str | None


class LogsShareData(BaseModel):
    id: str
    url: str
    raw: str


class RamData(BaseModel):
    ram: int


class MotdData(BaseModel):
    motd: str


class CommandData(BaseModel):
    command: str


class FileInfoData(BaseModel):
    path: str
    name: str
    isTextFile: bool
    isConfigFile: bool
    isDirectory: bool
    isLog: bool
    isReadable: bool
    isWritable: bool
    size: int
    children: list[str] | None


class ConfigFileType(Enum):
    string = auto()
    integer = auto()
    float = auto()
    boolean = auto()
    multiselect = auto()
    select = auto()


class ConfigFileData(BaseModel):
    key: str
    value: str
    label: str
    type: ConfigFileType
    options: list[str] | None = None


class CreditPoolData(BaseModel):
    id: str
    name: str
    credits: float
    servers: int
    owner: str
    isOwner: bool
    members: int
    ownShare: float
    ownCredits: float


class CreditPoolMemberData(BaseModel):
    account: str
    name: str
    share: int
    credits: float
    isOwner: bool





import asyncio

from typing import TYPE_CHECKING, Callable, Awaitable, assert_never, overload, Literal
from collections import defaultdict

import aiohttp

from aexaroton import BASE_URL
from aexaroton.socket_types import KeepAliveMessage, MessageType, ServerStatusMessage, SocketMessage, ReadyMessage, ConnectedMessage, DisconnectedMessage


if TYPE_CHECKING:
    from .server import Server


class ConsoleListener:
    def __init__(self, socket: "WebSocket") -> None:
        self.socket = socket

    async def send_command(self, command: str):
        await self.socket._send_console_command(command)

    async def process_line(self, line: str):
        raise NotImplementedError()


# TODO: add wait_for using futures
class WebSocket:
    def __init__(self, server: "Server"):
        self.server = server

        self.socket: aiohttp.ClientWebSocketResponse | None = None

        self._socket_task: asyncio.Task | None = None
        self._dispatch_tasks: list[asyncio.Task] = []

        self._event_listeners: defaultdict[MessageType, list[Callable[[SocketMessage], Awaitable[None]]]] = defaultdict(list)
        self._console_listener: ConsoleListener | None = None

        self._connecting_event = asyncio.Event()

    async def _is_connected(self) -> bool:
        return self.socket is not None

    async def connect(self):
        if self.socket is not None:
            raise ValueError("Called connect on an already connected websocket")
        
        if self._connecting_event.is_set():
            raise RuntimeError("Connecting event somehow set before connecting")

        self._socket_task = asyncio.create_task(self._socket_dispatch_task())
        await self._connecting_event.wait()

    async def close(self):
        if not await self._is_connected():
            raise ValueError("Cannot close an unopened socket")

        self.socket = None
        assert self._socket_task is not None
        self._socket_task.cancel()
        self._console_listener = None
        self._connecting_event.clear()

        await asyncio.gather(*self._dispatch_tasks)

    # TODO: how the heck do I type hint this
    async def register_listener(self, message_type: MessageType, callback: Callable[[SocketMessage], Awaitable[None]]):
        self._event_listeners[message_type].append(callback)

    async def register_console_listener(self, console_listener: type[ConsoleListener], *, tail_lines: int = 0):
        if not 0 <= tail_lines <= 500:
            raise ValueError(f"tail_lines must be between 0 and 500 (inclusive) not {tail_lines}")

        if self._console_listener is not None:
            raise ValueError("Must unregister previous console listener before registering a new one")

        if not await self._is_connected():
            raise ValueError("Socket must be open to register a console listener")

        await self._send(
            {
                "stream": "console",
                "type": "start",
                "data": {
                    "tail": tail_lines
                }
            }
        )

        self._console_listener = console_listener(self)

    async def unregister_console_listener(self):
        if not await self._is_connected():
            raise ValueError("Socket must be open to unregister a console listener")

        if self._console_listener is None:
            raise ValueError("There is no console listener to unregister")

        await self._send(
            {
                "stream": "console",
                "type": "stop"
            }
        )

        self._console_listener = None

    async def _send_console_command(self, command: str):
        if not await self._is_connected():
            raise ValueError("Socket must be open to send a console command")

        await self._send(
            {
                "stream": "console",
                "type": "command",
                "data": command
            }
        )

    async def _send(self, data: dict):
        if not await self._is_connected():
            raise RuntimeError("Somehow attempting to send in an unopened socket")

        assert self.socket is not None
        await self.socket.send_json(data)

    async def _socket_dispatch_task(self):
        async with self.server._session.ws_connect(BASE_URL / "servers" / self.server.data.id / "websocket") as ws:
            self.socket = ws
            self._connecting_event.set()

            async for message in ws:
                match message.type:
                    case aiohttp.WSMsgType.TEXT:
                        data: dict = message.json()
                        self._dispatch_tasks.append(asyncio.create_task(self._dispatch_event(data)))
                    case aiohttp.WSMsgType.ERROR | aiohttp.WSMsgType.closed:
                        break

    async def _dispatch_event(self, message: dict):
        if (stream_type := message.get("stream")) is not None:
            if stream_type == "status":
                pass
            else:
                # TODO: add other streams
                match stream_type:
                    case "console":
                        if self._console_listener is None:
                            raise RuntimeError("Somehow getting console stream without a console listener registered")

                        match message["type"]:
                            case "started":
                                pass
                            case "line":
                                await self._console_listener.process_line(message["data"])
                            case _:
                                raise RuntimeError(f"Unexpected message type in console stream {message['type']}")
                    case _:
                        raise NotImplementedError(f"Stream type {stream_type} not implemented")

                return

        message_type_name = message["type"].replace("-", "_")

        try:
            message_type = MessageType[message_type_name]
        except KeyError:
            raise RuntimeError(f"Unexpected message type {message_type_name}")

        match message_type:
            case MessageType.ready:
                event = ReadyMessage(**message)
            case MessageType.connected:
                event = ConnectedMessage(**message)
            case MessageType.disconnected:
                event = DisconnectedMessage(**message)
            case MessageType.keep_alive:
                event = KeepAliveMessage(**message)
            case MessageType.status:
                event = ServerStatusMessage(**message)
            case _:
                assert_never(message_type)

        await asyncio.gather(*[callback(event) for callback in self._event_listeners[message_type]])



from enum import IntEnum, auto, Enum

from pydantic import BaseModel

from aexaroton.types import ServerData


class MessageType(Enum):
    ready = "ready"
    connected = "connected"
    disconnected = "disconnected"
    keep_alive = "keep-alive"
    status = "status"


class ConsoleStreamType(IntEnum):
    start = auto()
    stop = auto()
    command = auto()
    started = auto()
    line = auto()


class TickStreamType(IntEnum):
    start = auto()
    stop = auto()
    started = auto()
    tick = auto()


class HeapStreamType(IntEnum):
    start = auto()
    stop = auto()
    started = auto()
    heap = auto()


class SocketMessage(BaseModel): ...


class ReadyMessage(SocketMessage):
    type: MessageType
    data: str


class ConnectedMessage(SocketMessage):
    type: MessageType


class DisconnectedMessage(SocketMessage):
    type: MessageType
    data: str


class KeepAliveMessage(SocketMessage):
    type: MessageType


class ServerStatusMessage(SocketMessage):
    type: MessageType
    stream: str
    data: ServerData



from __future__ import annotations
from typing import TYPE_CHECKING

from aexaroton import BASE_URL

from .types import ConfigFileData, FileInfoData, LogsData, LogsShareData, MotdData, RamData, ServerData, ExarotonResponse
from .websocket import WebSocket
from .errors import ExarotonError


if TYPE_CHECKING:
    from aiohttp import ClientSession


class Server:
    def __init__(self, data: ServerData, connection: "ClientSession"):
        self.data = data
        self._session = connection

    def __repr__(self) -> str:
        return f"<Server id={self.data.id}, name={self.data.name}>"

    async def create_socket(self) -> WebSocket:
        socket = WebSocket(self)
        await socket.connect()

        return socket

    async def get_logs(self) -> LogsData:
        async with self._session.get(BASE_URL / "servers" / self.data.id / "logs") as response:
            data: ExarotonResponse = await response.json()

            if data["success"] != True:
                raise ExarotonError(data["error"])

            assert data["data"] is not None
            return LogsData(**data["data"])

    async def share_logs(self) -> LogsShareData:
        async with self._session.get(BASE_URL / "servers" / self.data.id / "logs" / "share") as response:
            data: ExarotonResponse = await response.json()

            if data["success"] != True:
                raise ExarotonError(data["error"])

            assert data["data"] is not None
            return LogsShareData(**data["data"])

    async def get_ram(self) -> RamData:
        async with self._session.get(BASE_URL / "servers" / self.data.id / "options" / "ram") as response:
            data: ExarotonResponse = await response.json()

            if data["success"] != True:
                raise ExarotonError(data["error"])

            assert data["data"] is not None
            return RamData(**data["data"])

    async def set_ram(self, ram: int) -> RamData:
        async with self._session.post(BASE_URL / "servers" / self.data.id / "options" / "ram", json={"ram": ram}) as response:
            data: ExarotonResponse = await response.json()

            if data["success"] != True:
                raise ExarotonError(data["error"])

            assert data["data"] is not None
            return RamData(**data["data"])

    async def set_motd(self, motd: int) -> MotdData:
        async with self._session.post(BASE_URL / "servers" / self.data.id / "options" / "motd", json={"motd": motd}) as response:
            data: ExarotonResponse = await response.json()

            if data["success"] != True:
                raise ExarotonError(data["error"])

            assert data["data"] is not None
            return MotdData(**data["data"])

    async def start(self, use_own_credits: bool | None = None) -> str:
        if self.data.shared is False and use_own_credits is not None:
            raise ValueError("user_own_credits is only for shared servers")
        elif self.data.shared is True and use_own_credits is None:
            raise ValueError("use_own_credits must be set for shared servers")

        if self.data.shared is False:
            async with self._session.get(BASE_URL / "servers" / self.data.id / "start") as response:
                data: dict = await response.json()

                if data["success"] != True:
                    raise ExarotonError(data["error"])

                assert data["data"] is not None
                return data["data"]

        else:
            async with self._session.post(BASE_URL / "servers" / self.data.id / "start", json={"useOwnCredits": use_own_credits}) as response:
                data: dict = await response.json()

                if data["success"] != True:
                    raise ExarotonError(data["error"])

                assert data["data"] is not None
                return data["data"]

    async def stop(self) -> str:
        async with self._session.get(BASE_URL / "servers" / self.data.id / "stop") as response:
            data: dict = await response.json()

            if data["success"] != True:
                raise ExarotonError(data["error"])

            assert data["data"] is not None
            return data["data"]

    async def restart(self) -> str:
        async with self._session.get(BASE_URL / "servers" / self.data.id / "restart") as response:
            data: dict = await response.json()

            if data["success"] != True:
                raise ExarotonError(data["error"])

            assert data["data"] is not None
            return data["data"]

    async def command(self, command: str) -> str:
        async with self._session.post(BASE_URL / "servers" / self.data.id / "command", json={"command": command}) as response:
            data: dict = await response.json()

            return data["data"]

    async def get_player_lists(self) -> list[str]:
        async with self._session.get(BASE_URL / "servers" / self.data.id / "playerlists") as response:
            data: dict = await response.json()

            if data["success"] != True:
                raise ExarotonError(data["error"])

            assert data["data"] is not None
            return data["data"]

    # maybe there could be a PlayerList type?
    async def get_player_list(self, list_name: str) -> list[str]:
        async with self._session.get(BASE_URL / "servers" / self.data.id / "playerlists" / list_name) as response:
            data: dict = await response.json()

            if data["success"] != True:
                raise ExarotonError(data["error"])

            assert data["data"] is not None
            return data["data"]

    async def add_to_player_list(self, list_name: str, entries: list[str]) -> list[str]:
        async with self._session.put(BASE_URL / "servers" / self.data.id / "playerlists" / list_name, json={"entries": entries}) as response:
            data: dict = await response.json()

            if data["success"] != True:
                raise ExarotonError(data["error"])

            assert data["data"] is not None
            return data["data"]

    async def remove_from_player_list(self, list_name: str, entries: list[str]) -> list[str]:
        async with self._session.delete(BASE_URL / "servers" / self.data.id / "playerlists" / list_name, json={"entries": entries}) as response:
            data: dict = await response.json()

            if data["success"] != True:
                raise ExarotonError(data["error"])

            assert data["data"] is not None
            return data["data"]

    async def get_file_info(self, file_path: str) -> FileInfoData:
        async with self._session.get(BASE_URL / "servers" / self.data.id / "files" / "info" / file_path) as response:
            data: dict = await response.json()

            if data["success"] != True:
                raise ExarotonError(data["error"])

            assert data["data"] is not None
            return FileInfoData(**data["data"])

    async def get_file_data(self, file_path: str) -> bytes:
        async with self._session.get(BASE_URL / "servers" / self.data.id / "files" / "data" / file_path) as reponse:
            return await reponse.read()

    async def set_file_data(self, file_path: str, file_data: bytes) -> str:
        async with self._session.put(BASE_URL / "servers" / self.data.id / "files" / "data" / file_path, json=file_data) as response:
            data: dict = await response.json()

            if data["success"] != True:
                raise ExarotonError(data["error"])

            assert data["data"] is not None
            return data["data"]

    async def delete_file(self, file_path: str) -> str:
        async with self._session.delete(BASE_URL / "servers" / self.data.id / "files" / "data" / file_path) as response:
            data: dict = await response.json()

            if data["success"] != True:
                raise ExarotonError(data["error"])

            assert data["data"] is not None
            return data["data"]

    async def get_config_file(self, file_path: str) -> list[ConfigFileData]:
        async with self._session.get(BASE_URL / "servers" / self.data.id / "files" / "config" / file_path) as response:
            data: dict = await response.json()

            if data["success"] != True:
                raise ExarotonError(data["error"])

            assert data["data"] is not None
            return [ConfigFileData(**config_data) for config_data in data["data"]]

    async def set_config_file(self, file_path: str, options: dict[str, str]) -> list[ConfigFileData]:
        async with self._session.post(BASE_URL / "servers" / self.data.id / "files" / "config" / file_path, json=options) as response:
            data: dict = await response.json()

            if data["success"] != True:
                raise ExarotonError(data["error"])

            assert data["data"] is not None
            return [ConfigFileData(**config_data) for config_data in data["data"]]



from typing import Self

import aiohttp

from aexaroton import BASE_URL
from .types import AccountData, CreditPoolData, CreditPoolMemberData, ServerData, ExarotonResponse
from .server import Server
from .errors import ExarotonError


class Client:
    def __init__(self, token: str):
        self._closed: bool = False

        self.session = aiohttp.ClientSession(
            headers={
                "Authorization": f"Bearer {token}"
            }
        )

    async def __aenter__(self) -> Self:
        return self 

    async def __aexit__(self, *_) -> None:
        await self.close()
        return

    async def close(self):
        self._closed = True
        await self.session.close()

    async def get_account(self) -> AccountData:
        async with self.session.get(BASE_URL / "account") as response:
            data: dict = await response.json()

            return AccountData(**data["data"])

    async def get_servers(self) -> list[Server]:
        async with self.session.get(BASE_URL / "servers") as response:
            data: dict = await response.json()

            return [Server(ServerData(**server_data), self.session) for server_data in data["data"]]

    async def get_server(self, server_id: str) -> Server:
        async with self.session.get(BASE_URL / "servers" / server_id) as response:
            data: ExarotonResponse = await response.json()

            if data["success"] != True:
                raise ExarotonError(data["error"])

            assert data["data"] is not None
            return Server(ServerData(**data["data"]), self.session)

    async def get_credit_pools(self) -> list[CreditPoolData]:
        async with self.session.get(BASE_URL / "billing" / "pools") as response:
            data: ExarotonResponse = await response.json()

            if data["success"] != True:
                raise ExarotonError(data["error"])

            assert data["data"] is not None
            return [CreditPoolData(**pool_data) for pool_data in data["data"]]

    async def get_credit_pool(self, pool_id: str) -> CreditPoolData:
        async with self.session.get(BASE_URL / "billing" / "pools" / pool_id) as response:
            data: ExarotonResponse = await response.json()

            if data["success"] != True:
                raise ExarotonError(data["error"])

            assert data["data"] is not None
            return CreditPoolData(**data["data"])

    async def get_credit_pool_members(self, pool_id: str) -> list[CreditPoolMemberData]:
        async with self.session.get(BASE_URL / "billing" / "pools" / pool_id / "members") as response:
            data: ExarotonResponse = await response.json()

            if data["success"] != True:
                raise ExarotonError(data["error"])

            assert data["data"] is not None
            return [CreditPoolMemberData(**pool_member) for pool_member in data["data"]]

    async def get_credit_pool_servers(self, pool_id: str) -> list[Server]:
        async with self.session.get(BASE_URL / "billing" / "pools" / pool_id / "servers") as response:
            data: ExarotonResponse = await response.json()

            if data["success"] != True:
                raise ExarotonError(data["error"])

            assert data["data"] is not None
            return [Server(ServerData(**server_data), self.session) for server_data in data["data"]]


